@@ -39,6 +39,8 @@ class QWaylandInputDevice;
 class QWaylandXdgShell;
 class QWaylandXdgExportedV2;
 class QWaylandXdgExporterV2;
+class QWaylandXdgDialogWmV1;
+class QWaylandXdgDialogV1;
 
 class Q_WAYLANDCLIENT_EXPORT QWaylandXdgSurface : public QWaylandShellSurface, public QtWayland::xdg_surface
 {
@@ -62,11 +64,13 @@ public:
     void propagateSizeHints() override;
     void setWindowGeometry(const QRect &rect) override;
     bool requestActivate() override;
+    bool requestActivateOnShow() override;
     void setXdgActivationToken(const QString &token) override;
     void requestXdgActivationToken(quint32 serial) override;
     void setAlertState(bool enabled) override;
     bool isAlertState() const override { return m_alertState; }
     QString externWindowHandle() override;
+    void setWindowPosition(const QPoint &position) override;
 
     void setSizeHints();
 
@@ -101,6 +105,7 @@ private:
             QSize bounds = {0, 0};
             QSize size = {0, 0};
             Qt::WindowStates states = Qt::WindowNoState;
+            bool suspended = false;
         }  m_pending, m_applied;
         QWaylandWindow::ToplevelWindowTilingStates m_toplevelStates = QWaylandWindow::WindowNoState;
         QSize m_normalSize;
@@ -108,11 +113,18 @@ private:
         QWaylandXdgSurface *m_xdgSurface = nullptr;
         QWaylandXdgToplevelDecorationV1 *m_decoration = nullptr;
         QScopedPointer<QWaylandXdgExportedV2> m_exported;
+        QScopedPointer<QWaylandXdgDialogV1> m_xdgDialog;
+    };
+
+    class Positioner : public QtWayland::xdg_positioner {
+    public:
+        Positioner(QWaylandXdgShell *xdgShell);
+        ~Positioner() override;
     };
 
     class Popup : public QtWayland::xdg_popup {
     public:
-        Popup(QWaylandXdgSurface *xdgSurface, QWaylandWindow *parent, QtWayland::xdg_positioner *positioner);
+        Popup(QWaylandXdgSurface *xdgSurface, QWaylandWindow *parent, Positioner *positioner);
         ~Popup() override;
 
         void applyConfigure();
@@ -121,6 +133,7 @@ private:
         void grab(QWaylandInputDevice *seat, uint serial);
         void xdg_popup_configure(int32_t x, int32_t y, int32_t width, int32_t height) override;
         void xdg_popup_popup_done() override;
+        void xdg_popup_repositioned(uint32_t token) override;
 
         QWaylandXdgSurface *m_xdgSurface = nullptr;
         QWaylandXdgSurface *m_parentXdgSurface = nullptr;
@@ -128,18 +141,20 @@ private:
         bool m_grabbing = false;
 
         QRect m_pendingGeometry;
+        bool m_waitingForReposition = false;
+        uint32_t m_waitingForRepositionSerial = 0;
     };
 
     void setToplevel();
     void setPopup(QWaylandWindow *parent);
     void setGrabPopup(QWaylandWindow *parent, QWaylandInputDevice *device, int serial);
+    std::unique_ptr<Positioner> createPositioner(QWaylandWindow *parent);
 
     QWaylandXdgShell *m_shell = nullptr;
     QWaylandWindow *m_window = nullptr;
     Toplevel *m_toplevel = nullptr;
     Popup *m_popup = nullptr;
     bool m_configured = false;
-    QRegion m_exposeRegion;
     uint m_pendingConfigureSerial = 0;
     uint m_appliedConfigureSerial = 0;
     QString m_activationToken;
@@ -149,11 +164,11 @@ private:
     friend class QWaylandXdgShell;
 };
 
-class Q_WAYLANDCLIENT_EXPORT QWaylandXdgShell : public QtWayland::xdg_wm_base
+class Q_WAYLANDCLIENT_EXPORT QWaylandXdgShell
 {
 public:
-    QWaylandXdgShell(QWaylandDisplay *display, uint32_t id, uint32_t availableVersion);
-    ~QWaylandXdgShell() override;
+    QWaylandXdgShell(QWaylandDisplay *display, QtWayland::xdg_wm_base *xdg_wm_base);
+    ~QWaylandXdgShell();
 
     QWaylandDisplay *display() const { return m_display; }
 
@@ -162,18 +177,16 @@ public:
     QWaylandXdgExporterV2 *exporter() const { return m_xdgExporter.data(); }
     QWaylandXdgSurface *getXdgSurface(QWaylandWindow *window);
 
-protected:
-    void xdg_wm_base_ping(uint32_t serial) override;
-
 private:
     static void handleRegistryGlobal(void *data, ::wl_registry *registry, uint id,
                                      const QString &interface, uint version);
 
     QWaylandDisplay *m_display = nullptr;
+    QtWayland::xdg_wm_base *m_xdgWmBase = nullptr;
     QScopedPointer<QWaylandXdgDecorationManagerV1> m_xdgDecorationManager;
     QScopedPointer<QWaylandXdgActivationV1> m_xdgActivation;
     QScopedPointer<QWaylandXdgExporterV2> m_xdgExporter;
-    QWaylandXdgSurface::Popup *m_topmostGrabbingPopup = nullptr;
+    QScopedPointer<QWaylandXdgDialogWmV1> m_xdgDialogWm;
 
     friend class QWaylandXdgSurface;
 };
